#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>

#include <stdlib.h>
#include "threadp.h"

#include <iostream>
using namespace std;

#include "box_native.h"

// CRITICAL SECTIONS
bool boxValid_or_setValidAndStore_atomic(ClayBoxType box_record, unsigned long obj);
bool boxInvalid_or_setInvalid_atomic(ClayBoxType box_record);

class Box {

public:
	
	/**
	 * Box constructor
	 */
	Box()
		/* : cond_empty(PTHREAD_COND_INITIALIZER),
		  cond_full(PTHREAD_COND_INITIALIZER),
		  lock(PTHREAD_ERROR_CHECK_MUTEX_INITIALIZER_NP) */ {

		int rc = 0;
		rc += pthread_cond_init(&cond_empty, NULL);
		rc += pthread_cond_init(&cond_full, NULL);
		rc += pthread_mutex_init(&lock, NULL);
		// rc += pthread_mutex_init(&lock_cond, NULL);

		if (rc != 0) {
			cerr << "FATAL ERROR: Could not create synchronization primitives!";
			exit(-1);
		}

		valid = 0;
		ptr = 0;
	}

public:
	int valid; // boolean value
	long unsigned int ptr;

	// condition vars... possibly two here. 
	// one for signaling that the box is now empty for PUT OBJECT action
	pthread_cond_t cond_empty;
	// one for signaling that the box is now full for the GET OBJECT action
	pthread_cond_t cond_full;

	// lock
	// pthread_mutex_t lock_cond;
	pthread_mutex_t lock;
};


/**
 * create our box object
 */
ClayBoxType create_box() {

	ClayBoxType box_record_ptr;
	Box* box = new Box();

	// point to our native box
	box_record_ptr = (unsigned long)box;

	return box_record_ptr;
}


/**
 * create our box object
 */
ClayBoxType create_box_with_object(unsigned long value) {

	ClayBoxType box_record_ptr;
	Box* box = new Box();
	box->valid = 1;
	box->ptr = value;

	// point to our native box
	box_record_ptr = (unsigned long)box;

	return box_record_ptr;
}


bool box_valid(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	
	return box->valid == 1;
}


void box_validate(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	box->valid = 1;
}


void box_invalidate(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	box->valid = 0;
}


void box_lock(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	acquire(&box->lock, "Box Lock");
}


void box_unlock(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	release(&box->lock, "Box Unlock");
}


void box_wait_empty(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	cond_wait(&box->cond_empty, &box->lock, "Box CondWaitEmpty");
}


void box_wait_full(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	cond_wait(&box->cond_full, &box->lock, "Box CondWaitFull");
}


void box_signal_empty(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	cond_signal(&box->cond_empty, "Box CondSignalEmpty");
}


void box_signal_full(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	cond_signal(&box->cond_full, "Box CondSignalFull");
}


unsigned long int box_get(ClayBoxType box_record) {

	Box* box = (Box*)box_record;
	return box->ptr;
}


void box_set(ClayBoxType box_record, unsigned long value) {

	Box* box = (Box*)box_record;
	box->ptr = value;
}


/**
 * PUT OBJECT.
 */
void box_object(ClayBoxType box_record, unsigned long obj) {

	box_lock(box_record);
	
	do {
		// if we already have an object here... wait
		// till someone unboxes it out.
		if (box_valid(box_record)) {
			// wait till empty
			box_wait_empty(box_record);
		}
			
	} while (boxValid_or_setValidAndStore_atomic(box_record, obj));

	box_unlock(box_record);
}


/**
 * GET OBJECT.
 */
unsigned long unbox_object(ClayBoxType box_record) {

	unsigned long value;

	box_lock(box_record);

	do {
		// if we don't have any object here...
		if (!box_valid(box_record)) {
			// we wait till full
			box_wait_full(box_record);
		}

		value = box_get(box_record);

	} while (boxInvalid_or_setInvalid_atomic(box_record));

	box_unlock(box_record);
	
	return value;
}


/**
 * this action schould be atomic
 * called from box_object
 */
bool boxValid_or_setValidAndStore_atomic(ClayBoxType box_record, unsigned long obj) {

	//box_lock(box_record);
	
	if (box_valid(box_record)) {
		//box_unlock(box_record);
		return true;
	}

	box_validate(box_record);

	// store the object in box
	box_set(box_record, obj);

	// could signal here that object is ready for a GET OBJECT
	box_signal_full(box_record);

	//box_unlock(box_record);

	return false;
}


/**
 * this action should be atomic
 * called from unbox_object
 */
bool boxInvalid_or_setInvalid_atomic(ClayBoxType box_record) {

	//box_lock(box_record);

	if (!box_valid(box_record)) {

		// box_unlock(box_record);
		return true;
	}

	box_invalidate(box_record);

	box_signal_empty(box_record);

	//box_unlock(box_record);

	return false;
}
