#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>

#include <stdlib.h>
#include "threadp.h"

// local global
int verbose  = 0;

//-------------------------------------------------------------
// to make the calls look simpler...and to allow for verbose prints

void
acquire(pthread_mutex_t *mutexp, char *name) {
  int rc;

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: trying to acquire the lock)\n", name);
  }

  rc = pthread_mutex_lock(mutexp);

  if (rc) {
    fprintf(stderr,"%d hey, acquire failed!\n", rc);
    exit(-1);
  }

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: got the lock)\n", name);
  }

}

void
release(pthread_mutex_t *mutexp, char *name) {
  int rc;

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: releasing the lock)\n", name);
  }

  rc = pthread_mutex_unlock(mutexp);

  if (rc) {
    fprintf(stderr,"hey, release failed!\n");
    exit(-1);
  }
}


void
sem_up(sem_t *semp, char *name) {
  int rc;

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: upping the semaphore)\n", name);
  }

  rc = sem_post(semp);
  if (rc) {
    fprintf(stderr,"hey, sem_up failed!\n");
    exit(-1);
  }


}

void
sem_down(sem_t *semp, char *name) {
  int rc;

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: downing the semaphore)\n", name);
  }

  rc = sem_wait(semp); 
  // rc is always 0, see man page

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: through the semaphore)\n", name);
  }

}

int
sem_trydown(sem_t *semp, char *name) {
  int rc;

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: trying to down the semaphore)\n", name);
  }

  rc = sem_trywait(semp); 

  if (rc != 0) {
	  //printf("%s: Unable to Sem_down this time\n", name);
	  return 0;
  }

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: through the semaphore)\n", name);
  }

  return 1;
}

void
cond_wait(pthread_cond_t *cp, pthread_mutex_t *mp, char *name) {
  int rc;

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: atomically releasing the lock and waiting on the cvar)\n", 
		   name);
  }

  rc = pthread_cond_wait(cp,mp);
  // rc is always 0, see man page

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: woke up, and has the lock)\n", name);
  }

}


void
cond_signal(pthread_cond_t *cp, char *name) {
  int rc;

  if (verbose) {
	printf("%x ", (unsigned int)pthread_self());
    printf(".....(%s: signalling the cvar)\n",  name);
  }

  rc = pthread_cond_signal(cp);
  // rc is always 0, see man page

}

void
cond_broadcast(pthread_cond_t *cp, char *name) {
	int rc;

	if (verbose) {
		printf("%x ", (unsigned int)pthread_self());
		printf(".....(%s: broadcasting the cvar)\n",  name);
	}

	rc = pthread_cond_broadcast(cp);
	// rc is always 0, see man page

}

